@@ -3,7 +3,6 @@ from contextlib import contextmanager
 import os
 import re
 import sys
-import math
 import warnings
 import unittest
 from fnmatch import fnmatchcase
@@ -17,13 +16,11 @@ except ImportError:
     from collections import Iterable
 
 import numbers
-import json
-import importlib
 
 import numpy as np
-import openmdao
 
-from openmdao.core.constants import INT_DTYPE
+from openmdao.core.constants import INT_DTYPE, INF_BOUND
+from openmdao.utils.om_warnings import issue_warning, _warn_simple_format, warn_deprecation
 
 # Certain command line tools can make use of this to allow visualization of models when errors
 # are present that would normally cause setup to abort.
@@ -63,7 +60,7 @@ def ignore_errors(flag=None):
     Returns
     -------
     bool
-        The current value of _ignore_errors
+        The current value of _ignore_errors.
     """
     global _ignore_errors
     if flag is not None:
@@ -71,7 +68,7 @@ def ignore_errors(flag=None):
     return _ignore_errors
 
 
-def conditional_error(msg, exc=RuntimeError, category=UserWarning):
+def conditional_error(msg, exc=RuntimeError, category=UserWarning, err=None):
     """
     Raise an exception or issue a warning, depending on the value of _ignore_errors.
 
@@ -79,13 +76,16 @@ def conditional_error(msg, exc=RuntimeError, category=UserWarning):
     ----------
     msg : str
         The error/warning message.
-    exc : exception class
+    exc : Exception class
         This exception class is used to create the exception to be raised.
     category : warning class
         This category is the class of warning to be issued.
+    err : bool
+        If None, use ignore_errors(), otherwise use value of err to determine whether to
+        raise an exception (err=True) or issue a warning (err=False).
     """
-    if ignore_errors():
-        simple_warning(msg, category=category)
+    if (err is None and ignore_errors()) or err is False:
+        issue_warning(msg, category=category)
     else:
         raise exc(msg)
 
@@ -99,32 +99,17 @@ def ignore_errors_context(flag=True):
     ----------
     flag : bool
         If not None, set ignore_errors to this value.
+
+    Yields
+    ------
+    None
     """
     save = ignore_errors()
     ignore_errors(flag)
-    yield
-    ignore_errors(save)
-
-
-def warn_deprecation(msg):
-    """
-    Raise a warning and prints a deprecation message to stdout.
-
-    Parameters
-    ----------
-    msg : str
-        Message that will be printed to stdout.
-    """
-    # Deprecation warnings need to be printed regardless of debug level
-    warnings.simplefilter('always', DeprecationWarning)
-
-    # note, stack level 3 should take us back to original caller.
-    simple_warning(msg, DeprecationWarning, stacklevel=3)
-    warnings.simplefilter('ignore', DeprecationWarning)
-
-
-def _warn_simple_format(message, category, filename, lineno, file=None, line=None):
-    return '%s:%s: %s:%s\n' % (filename, lineno, category.__name__, message)
+    try:
+        yield
+    finally:
+        ignore_errors(save)
 
 
 def simple_warning(msg, category=UserWarning, stacklevel=2):
@@ -140,6 +125,8 @@ def simple_warning(msg, category=UserWarning, stacklevel=2):
     stacklevel : int
         Number of levels up the stack to identify as the warning location.
     """
+    warn_deprecation('simple_warning is deprecated. '
+                     'Use openmdao.utils.om_warnings.issue_warning instead.')
     old_format = warnings.formatwarning
     warnings.formatwarning = _warn_simple_format
     try:
@@ -148,93 +135,6 @@ def simple_warning(msg, category=UserWarning, stacklevel=2):
         warnings.formatwarning = old_format
 
 
-class reset_warning_registry(object):
-    """
-    Context manager which archives & clears warning registry for duration of context.
-
-    From https://bugs.python.org/file40031/reset_warning_registry.py
-
-    Attributes
-    ----------
-    _pattern : regex pattern
-        Causes manager to only reset modules whose names match this pattern. defaults to ``".*"``.
-    """
-
-    #: regexp for filtering which modules are reset
-    _pattern = None
-
-    #: dict mapping module name -> old registry contents
-    _backup = None
-
-    def __init__(self, pattern=None):
-        """
-        Initialize all attributes.
-
-        Parameters
-        ----------
-        pattern : regex pattern
-            Causes manager to only reset modules whose names match pattern. defaults to ``".*"``.
-        """
-        self._pattern = re.compile(pattern or ".*")
-
-    def __enter__(self):
-        """
-        Enter the runtime context related to this object.
-
-        Returns
-        -------
-        reset_warning_registry
-            This context manager.
-
-        """
-        # archive and clear the __warningregistry__ key for all modules
-        # that match the 'reset' pattern.
-        pattern = self._pattern
-        backup = self._backup = {}
-        for name, mod in list(sys.modules.items()):
-            if pattern.match(name):
-                reg = getattr(mod, "__warningregistry__", None)
-                if reg:
-                    backup[name] = reg.copy()
-                    reg.clear()
-        return self
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        """
-        Exit the runtime context related to this object.
-
-        Parameters
-        ----------
-        exc_type : Exception class
-            The type of the exception.
-        exc_value : Exception instance
-            The exception instance raised.
-        traceback : regex pattern
-            Traceback object.
-        """
-        # restore warning registry from backup
-        modules = sys.modules
-        backup = self._backup
-        for name, content in backup.items():
-            mod = modules.get(name)
-            if mod is None:
-                continue
-            reg = getattr(mod, "__warningregistry__", None)
-            if reg is None:
-                setattr(mod, "__warningregistry__", content)
-            else:
-                reg.clear()
-                reg.update(content)
-
-        # clear all registry entries that we didn't archive
-        pattern = self._pattern
-        for name, mod in list(modules.items()):
-            if pattern.match(name) and name not in backup:
-                reg = getattr(mod, "__warningregistry__", None)
-                if reg:
-                    reg.clear()
-
-
 def ensure_compatible(name, value, shape=None, indices=None):
     """
     Make value compatible with the specified shape or the shape of indices.
@@ -247,9 +147,8 @@ def ensure_compatible(name, value, shape=None, indices=None):
         The value of a variable.
     shape : int or tuple or list or None
         The expected or desired shape of the value.
-    indices : int or list of ints or tuple of ints or int ndarray or None
-        The indices of a source variable, used to determine shape if shape is None.
-        If shape is not None, the shape of the indices must match shape.
+    indices : Indexer or None
+        The indices into a source variable.
 
     Returns
     -------
@@ -267,14 +166,6 @@ def ensure_compatible(name, value, shape=None, indices=None):
     if isinstance(value, Iterable):
         value = np.asarray(value)
 
-    if indices is not None:
-        contains_slicer = _is_slicer_op(indices)
-        if not contains_slicer:
-            indices = np.atleast_1d(np.asarray(indices, dtype=INT_DTYPE))
-            ind_shape = indices.shape
-    else:
-        contains_slicer = False
-
     # if shape is not given, infer from value (if not scalar) or indices
     if shape is not None:
         if isinstance(shape, numbers.Integral):
@@ -283,14 +174,21 @@ def ensure_compatible(name, value, shape=None, indices=None):
             shape = tuple(shape)
     elif not np.isscalar(value):
         shape = np.atleast_1d(value).shape
-    elif indices is not None:
-        if len(ind_shape) > 1:
+
+    if indices is not None:
+        if not indices._flat_src and shape is None:
             raise RuntimeError("src_indices for '%s' is not flat, so its input "
-                               "shape must be provided. src_indices may contain "
-                               "an extra dimension if the connected source is "
-                               "not flat, making the input shape ambiguous." %
-                               name)
-        shape = ind_shape
+                               "shape must be provided." % name)
+        try:
+            indshape = indices.indexed_src_shape
+        except (RuntimeError, ValueError, TypeError):
+            pass  # use shape provided or shape of value and check vs. shape of indices later
+        else:
+            if shape is not None and np.product(indshape) != np.product(shape):
+                raise ValueError("Shape of indices %s does not match shape of %s for '%s'." %
+                                 (indshape, shape, name))
+            if shape is None:
+                shape = indshape
 
     if shape is None:
         # shape is not determined, assume the shape of value was intended
@@ -304,16 +202,10 @@ def ensure_compatible(name, value, shape=None, indices=None):
         else:
             value = np.atleast_1d(value).astype(np.float64)
             if value.shape != shape:
-                raise ValueError("Incompatible shape for '%s': "
-                                 "Expected %s but got %s." %
+                raise ValueError("Incompatible shape for '%s': Expected %s but got %s." %
                                  (name, shape, value.shape))
 
-    if indices is not None and not contains_slicer and shape != ind_shape[:len(shape)]:
-        raise ValueError("Shape of indices does not match shape for '%s': "
-                         "Expected %s but got %s." %
-                         (name, shape, ind_shape[:len(shape)]))
-
-    return value, shape, indices
+    return value, shape
 
 
 def determine_adder_scaler(ref0, ref, adder, scaler):
@@ -325,10 +217,10 @@ def determine_adder_scaler(ref0, ref, adder, scaler):
 
     Parameters
     ----------
-    ref : float or ndarray, optional
-        Value of response variable that scales to 1.0 in the driver.
     ref0 : float or ndarray, optional
         Value of response variable that scales to 0.0 in the driver.
+    ref : float or ndarray, optional
+        Value of response variable that scales to 1.0 in the driver.
     adder : float or ndarray, optional
         Value to add to the model value to get the scaled value. Adder
         is first in precedence.
@@ -339,7 +231,7 @@ def determine_adder_scaler(ref0, ref, adder, scaler):
     Returns
     -------
     tuple
-        adder and scaler, properly formatted and based on ref/ref0 if provided.
+        Adder and scaler, properly formatted and based on ref/ref0 if provided.
 
     Raises
     ------
@@ -391,14 +283,14 @@ def set_pyoptsparse_opt(optname, fallback=True):
     optname : str
         Name of pyoptsparse optimizer that is requested by the test.
     fallback : bool
-        If True, fall back to SLSQP if optname can't be found
+        If True, fall back to SLSQP if optname can't be found.
 
     Returns
     -------
     object
         Pyoptsparse optimizer instance.
     str
-        Pyoptsparse optimizer string
+        Pyoptsparse optimizer string.
     """
     OPT = None
     opt = None
@@ -407,10 +299,7 @@ def set_pyoptsparse_opt(optname, fallback=True):
     if force:
         optname = force
 
-    try:
-        from mock import Mock
-    except ImportError:
-        Mock = None
+    from unittest.mock import Mock
 
     try:
         from pyoptsparse import OPT
@@ -426,7 +315,7 @@ def set_pyoptsparse_opt(optname, fallback=True):
                 except Exception:
                     pass
         else:
-            if fallback and Mock and isinstance(opt, Mock):
+            if fallback and isinstance(opt, Mock):
                 try:
                     opt = OPT('SLSQP')
                     OPTIMIZER = 'SLSQP'
@@ -435,7 +324,7 @@ def set_pyoptsparse_opt(optname, fallback=True):
     except Exception:
         pass
 
-    if Mock and isinstance(opt, Mock):
+    if isinstance(opt, Mock):
         OPT = OPTIMIZER = None
 
     if not fallback and OPTIMIZER != optname:
@@ -488,9 +377,9 @@ def format_as_float_or_array(name, values, val_if_none=0.0, flatten=False):
     elif values is None:
         values = val_if_none
     elif values == float('inf'):
-        values = openmdao.INF_BOUND
+        values = INF_BOUND
     elif values == -float('inf'):
-        values = -openmdao.INF_BOUND
+        values = -INF_BOUND
     elif isinstance(values, numbers.Number):
         values = float(values)
     else:
@@ -533,7 +422,11 @@ def all_ancestors(pathname, delim='.'):
     pathname : str
         Pathname of starting object.
     delim : str
-        Delimiter used to split the name
+        Delimiter used to split the name.
+
+    Yields
+    ------
+    str
     """
     parts = pathname.split(delim)
     for i in range(len(parts), 0, -1):
@@ -547,7 +440,7 @@ def find_matches(pattern, var_list):
     Parameters
     ----------
     pattern : str
-        String pattern
+        String pattern.
     var_list : list of str
         List of variable names to search for pattern.
 
@@ -579,7 +472,7 @@ def pad_name(name, pad_num=10, quotes=False):
     Returns
     -------
     str
-        Padded string
+        Padded string.
     """
     l_name = len(name)
     quotes_len = 2 if quotes else 0
@@ -605,14 +498,14 @@ def run_model(prob, ignore_exception=False):
     Parameters
     ----------
     prob : Problem
-        an instance of Problem
+        An instance of Problem.
     ignore_exception : bool
         Set to True to ignore an exception of any kind.
 
     Returns
     -------
     string
-        output from calling `run_model` on the Problem, captured from stdout
+        Output from calling `run_model` on the Problem, captured from stdout.
     """
     stdout = sys.stdout
     strout = StringIO()
@@ -636,14 +529,14 @@ def run_driver(prob):
     Parameters
     ----------
     prob : Problem
-        an instance of Problem
+        An instance of Problem.
 
     Returns
     -------
-    boolean
+    bool
         Failure flag; True if failed to converge, False is successful.
     string
-        output from calling `run_driver` on the Problem, captured from stdout
+        Output from calling `run_driver` on the Problem, captured from stdout.
     """
     stdout = sys.stdout
     strout = StringIO()
@@ -666,15 +559,6 @@ def printoptions(*args, **kwds):
     options at the end. See `numpy.set_printoptions` for the full description of
     available options. If any invalid options are specified, they will be ignored.
 
-    Parameters
-    ----------
-    *args : list
-        Variable-length argument list.
-    **kwds : dict
-        Arbitrary keyword arguments.
-
-    Examples
-    --------
     >>> with printoptions(precision=2):
     ...     print(np.array([2.0])) / 3
     [0.67]
@@ -682,9 +566,20 @@ def printoptions(*args, **kwds):
     >>> with printoptions(precision=2) as opts:
     ...      assert_equal(opts, np.get_printoptions())
 
+    Parameters
+    ----------
+    *args : list
+        Variable-length argument list.
+    **kwds : dict
+        Arbitrary keyword arguments.
+
+    Yields
+    ------
+    str or int
+
     See Also
     --------
-    set_printoptions, get_printoptions
+        set_printoptions, get_printoptions
     """
     opts = np.get_printoptions()
 
@@ -699,6 +594,10 @@ def printoptions(*args, **kwds):
         np.set_printoptions(**opts)
 
 
+def _nothing():
+    yield None
+
+
 def do_nothing_context():
     """
     Do nothing.
@@ -711,88 +610,7 @@ def do_nothing_context():
     contextmanager
         A do nothing context manager.
     """
-    def nothing():
-        yield None
-
-    return contextmanager(nothing)()
-
-
-def _byteify(data, ignore_dicts=False):
-    """
-    Convert any unicode items in a data structure to bytes (object_hook for json load/loads).
-
-    Credit: Mirec Miskuf
-    stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-from-json
-
-    Parameters
-    ----------
-    data : any data item or structure
-        the data to be converted
-    ignore_dicts : bool
-        a flag to prevent recursion on dicts that have already been byteified.
-        False when object_hook passes a new dict to byteify, True at all other times.
-
-    Returns
-    -------
-    data item or structure
-        data item or structure with unicode converted to bytes
-    """
-    # if this is a unicode string, return its string representation
-    if isinstance(data, unicode):
-        return data.encode('utf-8')
-
-    # if this is a list of values, return list of byteified values
-    if isinstance(data, list):
-        return [_byteify(item, ignore_dicts=True) for item in data]
-
-    # if this is a dictionary, return dictionary of byteified keys and values
-    # but only if we haven't already byteified it
-    if isinstance(data, dict) and not ignore_dicts:
-        return {
-            _byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True)
-            for key, value in data.iteritems()
-        }
-
-    # if it's anything else, return it in its original form
-    return data
-
-
-def json_load_byteified(file_handle):
-    """
-    Load data from a JSON file, converting unicode to bytes if Python version is 2.x.
-
-    Intended for use only with Python 2.x, behaves the same as json.load() under Python 3.x.
-
-    Parameters
-    ----------
-    file_handle : file
-        file containing the data to be converted
-
-    Returns
-    -------
-    data item or structure
-        data item or structure with unicode converted to bytes
-    """
-    return json.load(file_handle)
-
-
-def json_loads_byteified(json_str):
-    """
-    Load data from a JSON string, converting unicode to bytes if Python version is 2.x.
-
-    Intended for use only with Python 2.x, behaves the same as json.loads() under Python 3.x.
-
-    Parameters
-    ----------
-    json_str : str
-        text string containing json encoded data
-
-    Returns
-    -------
-    data item or structure
-        data item or structure with unicode converted to bytes
-    """
-    return json.loads(json_str)
+    return contextmanager(_nothing)()
 
 
 def remove_whitespace(s, right=False, left=False):
@@ -859,7 +677,7 @@ def make_serializable(o):
     Parameters
     ----------
     o : object
-        the object to be converted
+        The object to be converted.
 
     Returns
     -------
@@ -878,10 +696,13 @@ def make_serializable(o):
         return o.item()
     elif isinstance(o, (str, float, int)):
         return o
-    elif isinstance(o, bool) or np.iscomplex(o):
+    elif isinstance(o, bool) or isinstance(o, complex):
         return str(o)
     elif hasattr(o, '__dict__'):
-        return o.__class__.__name__
+        try:
+            return o.to_json()
+        except AttributeError:
+            return o.__class__.__name__
     else:
         return o
 
@@ -896,7 +717,7 @@ def make_serializable_key(o):
     Parameters
     ----------
     o : object
-        the object to be converted
+        The object to be converted.
 
     Returns
     -------
@@ -924,7 +745,7 @@ def default_noraise(o):
     Parameters
     ----------
     o : object
-        the object to be converted
+        The object to be converted.
 
     Returns
     -------
@@ -943,7 +764,7 @@ def default_noraise(o):
         return o.item()
     elif isinstance(o, (str, float, int)):
         return o
-    elif isinstance(o, bool) or np.iscomplex(o):
+    elif isinstance(o, bool) or isinstance(o, complex):
         return str(o)
     elif hasattr(o, '__dict__'):
         return o.__class__.__name__
@@ -1125,7 +946,7 @@ def common_subpath(pathnames):
 
 def _is_slicer_op(indices):
     """
-    Check if an array of indices contains a colon or ellipsis operator.
+    Check if an indexer contains a slice or ellipsis operator.
 
     Parameters
     ----------
@@ -1161,10 +982,65 @@ def _slice_indices(slicer, arr_size, arr_shape):
     array
         Returns the sliced indices.
     """
-    return np.arange(arr_size, dtype=INT_DTYPE).reshape(arr_shape)[slicer]
+    if isinstance(slicer, slice):
+        # for a simple slice we can use less memory
+        start, stop, step = slicer.start, slicer.stop, slicer.step
+        if start is None:
+            start = 0
+        if stop is None:
+            stop = arr_size
+        if step is None:
+            step = 1
+        return np.arange(start, stop, step, dtype=INT_DTYPE).reshape(arr_shape)
+    else:
+        return np.arange(arr_size, dtype=INT_DTYPE).reshape(arr_shape)[slicer]
 
 
-def prom2ivc_src_dict(prom_dict):
+def _prom2ivc_src_name_iter(prom_dict):
+    """
+    Yield keys from prom_dict with promoted input names converted to ivc source names.
+
+    Parameters
+    ----------
+    prom_dict : dict
+        Original dict with some promoted paths.
+
+    Yields
+    ------
+    str
+        name
+    """
+    for name, meta in prom_dict.items():
+        if meta['ivc_source'] is not None:
+            yield meta['ivc_source']
+        else:
+            yield name
+
+
+def _prom2ivc_src_item_iter(prom_dict):
+    """
+    Yield items from prom_dict with promoted input names converted to ivc source names.
+
+    The result is that all names are absolute.
+
+    Parameters
+    ----------
+    prom_dict : dict
+        Original dict with some promoted paths.
+
+    Yields
+    ------
+    tuple
+        name, metadata
+    """
+    for name, meta in prom_dict.items():
+        if meta['ivc_source'] is not None:
+            yield meta['ivc_source'], meta
+        else:
+            yield name, meta
+
+
+def _prom2ivc_src_dict(prom_dict):
     """
     Convert a dictionary with promoted input names into one with ivc source names.
 
@@ -1178,15 +1054,7 @@ def prom2ivc_src_dict(prom_dict):
     dict
         New dict with ivc source pathnames.
     """
-    src_dict = {}
-    for name, meta in prom_dict.items():
-        if meta['ivc_source'] is not None:
-            src_name = meta['ivc_source']
-            src_dict[src_name] = meta
-        else:
-            src_dict[name] = meta
-
-    return src_dict
+    return {name: meta for name, meta in _prom2ivc_src_item_iter(prom_dict)}
 
 
 def convert_src_inds(parent_src_inds, parent_src_shape, my_src_inds, my_src_shape):
@@ -1200,7 +1068,7 @@ def convert_src_inds(parent_src_inds, parent_src_shape, my_src_inds, my_src_shap
     parent_src_shape : tuple
         Shape of source expected by parent.
     my_src_inds : ndarray or fancy index
-        src_indices at the current system level, before conversion.
+        Src_indices at the current system level, before conversion.
     my_src_shape : tuple
         Expected source shape at the current system level.
 
@@ -1213,89 +1081,11 @@ def convert_src_inds(parent_src_inds, parent_src_shape, my_src_inds, my_src_shap
         return my_src_inds
     elif my_src_inds is None:
         return parent_src_inds
-    if isinstance(my_src_inds, tuple):
-        ndims = len(my_src_inds)
-    elif isinstance(my_src_inds, np.ndarray):
-        ndims = my_src_inds.ndim
-    else:  # slice
-        ndims = 1
-    if _is_slicer_op(parent_src_inds):
-        parent_src_inds = _slice_indices(parent_src_inds, np.prod(parent_src_shape),
-                                         parent_src_shape)
-    if ndims == 1:
-        return parent_src_inds.ravel()[my_src_inds]
+
+    if my_src_inds._flat_src:
+        return parent_src_inds.shaped_array(flat=True)[my_src_inds.flat()]
     else:
-        return parent_src_inds.reshape(my_src_shape)[my_src_inds]
-
-
-def shape_from_idx(src_shape, src_inds, flat_src_inds):
-    """
-    Get the shape of the result if the given src_inds were applied to an array of the given shape.
-
-    Parameters
-    ----------
-    src_shape : tuple
-        Expected shape of source variable.
-    src_inds : ndarray or fancy index
-        Indices into the source variable.
-    flat_src_inds : bool
-        If True, src_inds index into a flat array.
-
-    Returns
-    -------
-    tuple
-        Shape of the input.
-    """
-    if _is_slicer_op(src_inds):
-        if isinstance(src_inds, slice):
-            src_inds = [src_inds]
-
-        for entry in src_inds:
-            if entry is Ellipsis:
-                shp = np.empty(src_shape, dtype=bool)[src_inds].shape
-                if flat_src_inds:
-                    return (np.product(shp),)
-                return shp
-
-        if len(src_inds) != len(src_shape):
-            raise ValueError(f"src_indices {src_inds} have the wrong number of dimensions, "
-                             f"{len(src_inds)}, to index into an array of shape {src_shape}")
-
-        ret = []
-        full_slice = slice(None)
-        for i, dimsize in enumerate(src_shape):
-            slc = src_inds[i]
-            if isinstance(slc, slice):
-                if slc == full_slice:
-                    ret.append(dimsize)
-                elif (slc.start is None or slc.start < 0 or slc.stop is None or
-                      slc.stop < 0 or (slc.step is not None and slc.step < 0)):
-                    ret.append(np.empty(dimsize, dtype=bool)[slc].size)
-                else:
-                    step = slc.step
-                    if step in (None, 1):
-                        ret.append(slc.stop - slc.start)
-                    else:
-                        ret.append(1 + (slc.stop - slc.start - 1) // slc.step)
-            elif isinstance(slc, Number):
-                ret.append(1)
-            else:  # list/tuple/array
-                ret.append(len(slc))
-        return tuple(ret)
-    else:
-        if src_inds.ndim == 1:
-            flat_src_inds = True
-        if flat_src_inds:
-            return src_inds.shape
-        elif src_inds.size > 0:
-            dims = src_inds.shape[-1]
-        else:
-            return (0,)
-
-        if len(src_shape) != dims:
-            raise ValueError(f"non-flat src_indices {src_inds} have the wrong number of dimensions,"
-                             f" {dims}, to index into an array of shape {src_shape}")
-        return tuple(src_inds.shape[:-1])
+        return parent_src_inds.shaped_array(flat=False).reshape(my_src_shape)[my_src_inds()]
 
 
 def shape2tuple(shape):
@@ -1321,9 +1111,9 @@ def shape2tuple(shape):
 
 def get_connection_owner(system, tgt):
     """
-    Return (owner, promoted_src, promoted_tgt) for the given connected source and target.
+    Return (owner, promoted_src, promoted_tgt) for the given connected target.
 
-    Note: this is not speedy.  It's intended for use only in error messages.
+    Note : this is not speedy.  It's intended for use only in error messages.
 
     Parameters
     ----------
@@ -1335,19 +1125,144 @@ def get_connection_owner(system, tgt):
     Returns
     -------
     tuple
-        (owning group, promoted source name, promoted target name)
+        (wning group, promoted source name, promoted target name).
     """
     from openmdao.core.group import Group
 
     model = system._problem_meta['model_ref']()
     src = model._conn_global_abs_in2out[tgt]
+    abs2prom = model._var_allprocs_abs2prom
 
-    if model._var_allprocs_abs2prom['input'][tgt] != model._var_allprocs_abs2prom['output'][src]:
-        # connection is explicit
-        for g in model.system_iter(include_self=True, recurse=True, typ=Group):
-            if g._manual_connections:
-                tprom = g._var_allprocs_abs2prom['input'][tgt]
-                if tprom in g._manual_connections:
-                    return g.pathname, g._var_allprocs_abs2prom['output'][src], tprom
+    if src in abs2prom['output'] and tgt in abs2prom['input'][tgt]:
+        if abs2prom['input'][tgt] != abs2prom['output'][src]:
+            # connection is explicit
+            for g in model.system_iter(include_self=True, recurse=True, typ=Group):
+                if g._manual_connections:
+                    tprom = g._var_allprocs_abs2prom['input'][tgt]
+                    if tprom in g._manual_connections:
+                        return g.pathname, g._var_allprocs_abs2prom['output'][src], tprom
 
     return None, None, None
+
+
+def wing_dbg():
+    """
+    Make import of wingdbstub contingent on value of WING_DBG environment variable.
+
+    Also will import wingdbstub from the WINGHOME directory.
+    """
+    if env_truthy('WING_DBG'):
+        import sys
+        import os
+        save = sys.path
+        new = sys.path[:] + [os.environ['WINGHOME']]
+        sys.path = new
+        try:
+            import wingdbstub
+        finally:
+            sys.path = save
+
+
+class LocalRangeIterable(object):
+    """
+    Iterable object yielding local indices while iterating over local or distributed vars.
+
+    The number of iterations for a distributed variable will be the full distributed size of the
+    variable but None will be returned for any indices that are not local to the given rank.
+
+    Parameters
+    ----------
+    system : System
+        Containing System.
+    vname : str
+        Name of the variable.
+    use_vec_offset : bool
+        If True, return indices for the given variable within its vector, else just return
+        indices within the variable itself, i.e. range(var_size).
+
+    Attributes
+    ----------
+    _inds : ndarray
+        Variable indices (unused for distributed variables).
+    _dist_size : int
+        Full size of distributed variable.
+    _start : int
+        Starting index of distributed variable on this rank.
+    _end : int
+        Last index + 1 of distributed variable on this rank.
+    _offset : int
+        Offset of this variable into the local vector,.
+    _iter : method
+        The iteration method used.
+    """
+
+    def __init__(self, system, vname, use_vec_offset=True):
+        """
+        Initialize the iterator.
+        """
+        self._dist_size = 0
+
+        abs2meta = system._var_allprocs_abs2meta['output']
+        if vname in abs2meta:
+            sizes = system._var_sizes['output']
+            slices = system._outputs.get_slice_dict()
+        else:
+            abs2meta = system._var_allprocs_abs2meta['input']
+            sizes = system._var_sizes['input']
+            slices = system._inputs.get_slice_dict()
+
+        if abs2meta[vname]['distributed']:
+            var_idx = system._var_allprocs_abs2idx[vname]
+            rank = system.comm.rank
+            self._offset = np.sum(sizes[rank, :var_idx]) if use_vec_offset else 0
+
+            self._iter = self._dist_iter
+            self._start = np.sum(sizes[:rank, var_idx])
+            self._end = self._start + sizes[rank, var_idx]
+            self._dist_size = np.sum(sizes[:, var_idx])
+        else:
+            self._iter = self._serial_iter
+            if use_vec_offset:
+                self._inds = range(slices[vname].start, slices[vname].stop)
+            else:
+                self._inds = range(slices[vname].stop - slices[vname].start)
+
+    def _serial_iter(self):
+        """
+        Iterate over a local non-distributed variable.
+
+        Yields
+        ------
+        int
+            Variable index.
+        """
+        yield from self._inds
+
+    def _dist_iter(self):
+        """
+        Iterate over a distributed variable.
+
+        Yields
+        ------
+        int or None
+            Variable index or None if index is not local to this rank.
+        """
+        start = self._start
+        end = self._end
+
+        for i in range(self._dist_size):
+            if i >= start and i < end:
+                yield i - start + self._offset
+            else:
+                yield None
+
+    def __iter__(self):
+        """
+        Return an iterator.
+
+        Returns
+        -------
+        iterator
+            An iterator over our indices.
+        """
+        return self._iter()
