@@ -6,6 +6,7 @@ import weakref
 import numpy as np
 
 from openmdao.utils.name_maps import prom_name2abs_name, rel_name2abs_name
+from openmdao.utils.indexer import Indexer, indexer
 
 
 _full_slice = slice(None)
@@ -27,10 +28,23 @@ class Vector(object):
     - <DefaultVector>
     - <PETScVector>
 
+    Parameters
+    ----------
+    name : str
+        The name of the vector: 'nonlinear' or 'linear'.
+    kind : str
+        The kind of vector, 'input', 'output', or 'residual'.
+    system : <System>
+        Pointer to the owning system.
+    root_vector : <Vector>
+        Pointer to the vector owned by the root system.
+    alloc_complex : bool
+        Whether to allocate any imaginary storage to perform complex step. Default is False.
+
     Attributes
     ----------
     _name : str
-        The name of the vector: 'nonlinear', 'linear', or right-hand side name.
+        The name of the vector: 'nonlinear' or 'linear'.
     _typ : str
         Type: 'input' for input vectors; 'output' for output/residual vectors.
     _kind : str
@@ -49,65 +63,43 @@ class Vector(object):
         Set of variables that are relevant in the current context.
     _root_vector : Vector
         Pointer to the vector owned by the root system.
-    _alloc_complex : Bool
+    _alloc_complex : bool
         If True, then space for the complex vector is also allocated.
     _data : ndarray
         Actual allocated data.
     _slices : dict
         Mapping of var name to slice.
-    _cplx_data : ndarray
-        Actual allocated data under complex step.
-    _cplx_views : dict
-        Dictionary mapping absolute variable names to the ndarray views under complex step.
-    _cplx_views_flat : dict
-        Dictionary mapping absolute variable names to the flattened ndarray views under complex
-        step.
     _under_complex_step : bool
         When True, this vector is under complex step, and data is swapped with the complex data.
-    _ncol : int
-        Number of columns for multi-vectors.
-    _icol : int or None
-        If not None, specifies the 'active' column of a multivector when interfaceing with
-        a component that does not support multivectors.
     _do_scaling : bool
         True if this vector performs scaling.
+    _do_adder : bool
+        True if this vector's scaling includes an additive term.
     _scaling : dict
         Contains scale factors to convert data arrays.
+    _scaling_nl_vec : dict
+        Reference to the scaling factors in the nonlinear vector. Only used for linear input
+        vectors.
     read_only : bool
         When True, values in the vector cannot be changed via the user __setitem__ API.
-    _under_complex_step : bool
-        When True, self._data is replaced with self._cplx_data.
     _len : int
         Total length of data vector (including shared memory parts).
+    _has_solver_ref : bool
+        This is set to True only when a ref is defined on a solver.
     """
 
-    # Listing of relevant citations that should be referenced when
+    # Listing of relevant citations
     cite = ""
+    # Indicator whether a vector class is MPI-distributed
+    distributed = False
 
-    def __init__(self, name, kind, system, root_vector=None, alloc_complex=False, ncol=1):
+    def __init__(self, name, kind, system, root_vector=None, alloc_complex=False):
         """
         Initialize all attributes.
-
-        Parameters
-        ----------
-        name : str
-            The name of the vector: 'nonlinear', 'linear', or right-hand side name.
-        kind : str
-            The kind of vector, 'input', 'output', or 'residual'.
-        system : <System>
-            Pointer to the owning system.
-        root_vector : <Vector>
-            Pointer to the vector owned by the root system.
-        alloc_complex : bool
-            Whether to allocate any imaginary storage to perform complex step. Default is False.
-        ncol : int
-            Number of columns for multi-vectors.
         """
         self._name = name
         self._typ = _type_map[kind]
         self._kind = kind
-        self._ncol = ncol
-        self._icol = None
         self._len = 0
 
         self._system = weakref.ref(system)
@@ -125,16 +117,21 @@ class Vector(object):
 
         # Support for Complex Step
         self._alloc_complex = alloc_complex
-        self._cplx_data = None
-        self._cplx_views = {}
-        self._cplx_views_flat = {}
         self._under_complex_step = False
 
         self._do_scaling = ((kind == 'input' and system._has_input_scaling) or
                             (kind == 'output' and system._has_output_scaling) or
                             (kind == 'residual' and system._has_resid_scaling))
+        self._do_adder = ((kind == 'input' and system._has_input_adder) or
+                          (kind == 'output' and system._has_output_adder) or
+                          (kind == 'residual' and system._has_resid_scaling))
 
-        self._scaling = {}
+        self._scaling = None
+        self._scaling_nl_vec = None
+
+        # If we define 'ref' on an output, then we will need to allocate a separate scaling ndarray
+        # for the linear and nonlinear input vectors.
+        self._has_solver_ref = system._has_output_scaling and kind == 'input' and name == 'linear'
 
         if root_vector is None:
             self._root_vector = self
@@ -186,7 +183,7 @@ class Vector(object):
         Returns
         -------
         listiterator (Python 3.x) or list (Python 2.x)
-            the variable names.
+            The variable names.
         """
         return self.__iter__()
 
@@ -194,12 +191,19 @@ class Vector(object):
         """
         Return values of variables contained in this vector.
 
-        Returns
-        -------
-        list
-            the variable values.
+        Yields
+        ------
+        ndarray or float
+            Value of each variable.
         """
-        return [v for n, v in self._views.items() if n in self._names]
+        if self._under_complex_step:
+            for n, v in self._views.items():
+                if n in self._names:
+                    yield v
+        else:
+            for n, v in self._views_flat.items():
+                if n in self._names:
+                    yield v.real
 
     def _name2abs_name(self, name):
         """
@@ -230,7 +234,7 @@ class Vector(object):
 
     def __iter__(self):
         """
-        Yield an iterator over variables involved in the current mat-vec product (relative names).
+        Return an iterator over variables involved in the current mat-vec product (relative names).
 
         Returns
         -------
@@ -251,18 +255,32 @@ class Vector(object):
         ----------
         flat : bool
             If True, return the flattened values.
+
+        Yields
+        ------
+        str
+            Name of each variable.
+        ndarray or float
+            Value of each variable.
         """
         arrs = self._views_flat if flat else self._views
 
-        for name, val in arrs.items():
-            yield name, val
+        if self._under_complex_step:
+            yield from arrs.items()
+        else:
+            for name, val in arrs.items():
+                yield name, val.real
 
     def _abs_iter(self):
         """
         Iterate over the absolute names in the vector.
+
+        Yields
+        ------
+        str
+            Name of each variable.
         """
-        for name in self._views:
-            yield name
+        yield from self._views
 
     def __contains__(self, name):
         """
@@ -275,7 +293,7 @@ class Vector(object):
 
         Returns
         -------
-        boolean
+        bool
             True or False.
         """
         return self._name2abs_name(name) is not None
@@ -291,7 +309,7 @@ class Vector(object):
 
         Returns
         -------
-        boolean
+        bool
             True or False.
         """
         return name in self._names
@@ -312,10 +330,7 @@ class Vector(object):
         """
         abs_name = self._name2abs_name(name)
         if abs_name is not None:
-            if self._icol is None:
-                return self._views[abs_name]
-            else:
-                return self._views[abs_name][:, self._icol]
+            return self._abs_get_val(abs_name, flat=False)
         else:
             raise KeyError(f"{self._system().msginfo}: Variable name '{name}' not found.")
 
@@ -338,8 +353,13 @@ class Vector(object):
             variable value.
         """
         if flat:
-            return self._views_flat[name]
-        return self._views[name]
+            val = self._views_flat[name]
+        else:
+            val = self._views[name]
+
+        if self._under_complex_step:
+            return val
+        return val.real
 
     def __setitem__(self, name, value):
         """
@@ -428,25 +448,13 @@ class Vector(object):
         Parameters
         ----------
         val : int or float
-            scalar.
+            Scalar.
         vec : <Vector>
-            this vector times val is added to self.
+            This vector times val is added to self.
         """
         raise NotImplementedError('add_scale_vec not defined for vector type %s' %
                                   type(self).__name__)
 
-    def scale(self, scale_to):
-        """
-        Scale this vector to normalized or physical form.
-
-        Parameters
-        ----------
-        scale_to : str
-            Values are "phys" or "norm" to scale to physical or normalized.
-        """
-        raise NotImplementedError('scale not defined for vector type %s' %
-                                  type(self).__name__)
-
     def asarray(self, copy=False):
         """
         Return a flat array representation of this vector.
@@ -491,7 +499,7 @@ class Vector(object):
         Parameters
         ----------
         vec : <Vector>
-            the vector whose values self is set to.
+            The vector whose values self is set to.
         """
         raise NotImplementedError('set_vec not defined for vector type %s' %
                                   type(self).__name__)
@@ -505,8 +513,8 @@ class Vector(object):
         Parameters
         ----------
         val : float or ndarray
-            scalar or array to set data array to.
-        idxs : int or slice or tuple of ints and/or slices.
+            Scalar or array to set data array to.
+        idxs : int or slice or tuple of ints and/or slices
             The locations where the data array should be updated.
         """
         raise NotImplementedError('set_arr not defined for vector type %s' %
@@ -522,7 +530,7 @@ class Vector(object):
             The name of the variable.
         val : float or ndarray
             Scalar or array to set data array to.
-        idxs : int or slice or tuple of ints and/or slices.
+        idxs : int or slice or tuple of ints and/or slices
             The locations where the data array should be updated.
         flat : bool
             If True, set into flattened variable.
@@ -535,23 +543,31 @@ class Vector(object):
             raise ValueError(f"{self._system().msginfo}: Attempt to set value of '{name}' in "
                              f"{self._kind} vector when it is read only.")
 
-        if self._icol is not None:
-            idxs = (idxs, self._icol)
+        if not isinstance(idxs, Indexer):
+            idxs = indexer(idxs, flat_src=flat)
 
-        value = np.asarray(val)
-
-        try:
-            if flat:
-                self._views_flat[abs_name][idxs] = value.flat
+        if flat:
+            if isinstance(val, float):
+                self._views_flat[abs_name][idxs.flat()] = val
             else:
-                self._views[abs_name][idxs] = value
-        except Exception as err:
+                self._views_flat[abs_name][idxs.flat()] = np.asarray(val).flat
+        else:
+            value = np.asarray(val)
+            view = self._views[abs_name]
             try:
-                value = value.reshape(self._views[abs_name][idxs].shape)
-            except Exception:
-                raise ValueError(f"{self._system().msginfo}: Failed to set value of "
-                                 f"'{name}': {str(err)}.")
-            self._views[abs_name][idxs] = value
+                if view.shape:
+                    view[idxs()] = value
+                else:
+                    # view is a scalar (so not really a view), so set the value into the
+                    # array using the flat view (which is actually a view)
+                    self._views_flat[abs_name][0] = value
+            except Exception as err:
+                try:
+                    value = value.reshape(view[idxs()].shape)
+                except Exception:
+                    raise ValueError(f"{self._system().msginfo}: Failed to set value of "
+                                     f"'{name}': {str(err)}.")
+                view[idxs()] = value
 
     def dot(self, vec):
         """
@@ -576,7 +592,7 @@ class Vector(object):
         Returns
         -------
         float
-            norm of this vector.
+            Norm of this vector.
         """
         raise NotImplementedError('get_norm not defined for vector type %s' %
                                   type(self).__name__)
@@ -589,33 +605,13 @@ class Vector(object):
         raise NotImplementedError('_in_matvec_context not defined for vector type %s' %
                                   type(self).__name__)
 
-    def set_complex_step_mode(self, active, keep_real=False):
+    def set_complex_step_mode(self, active):
         """
         Turn on or off complex stepping mode.
 
-        When turned on, the default real ndarray is replaced with a complex ndarray and all
-        pointers are updated to point to it.
-
         Parameters
         ----------
         active : bool
             Complex mode flag; set to True prior to commencing complex step.
-
-        keep_real : bool
-            When this flag is True, keep the real value when turning off complex step. You only
-            need to do this when temporarily disabling complex step for guess_nonlinear.
         """
-        if active:
-            arr = self._data
-        elif keep_real:
-            arr = self._data.real
-        else:
-            arr = None
-
-        self._data, self._cplx_data = self._cplx_data, self._data
-        self._views, self._cplx_views = self._cplx_views, self._views
-        self._views_flat, self._cplx_views_flat = self._cplx_views_flat, self._views_flat
         self._under_complex_step = active
-
-        if arr is not None:
-            self.set_val(arr)
